from __future__ import annotations

from scipp import Variable
from mccode_antlr.common.parameters import InstrumentParameter, Expr


def is_type(x, t, name):
    if not isinstance(x, t):
        raise RuntimeError(f"{name} must be a {t}")


def has_compatible_unit(x: Variable, unit):
    from scipp import UnitError
    try:
        x.to(unit=unit, copy=False)
    except UnitError:
        return False
    return True


def is_scalar(x: Variable):
    from scipp import DimensionError
    try:
        y = x.value
    except DimensionError:
        return False
    return True


def variable_value_or_parameter(value: Variable | InstrumentParameter, unit: str):
    if isinstance(value, Variable):
        return value.to(unit=unit).value
    if isinstance(value, InstrumentParameter):
        return value.name
    return value


def serializable(a):
    if isinstance(a, dict):
        return serializable_dict(a)
    return serializable_one(a)


def serializable_dict(data: dict):
    for k, v in data.items():
        data[k] = serializable(v)
    return data


def serializable_one(v):
    if isinstance(v, Variable):
        return serialize_scipp(v)
    elif isinstance(v, InstrumentParameter) or isinstance(v, Expr):
        return serialize_mccode(v)
    elif isinstance(v, dict):
        return serializable_dict(v)
    elif isinstance(v, tuple):
        return tuple(serializable(a) for a in v)
    elif isinstance(v, list):
        return [serializable(a) for a in v]
    return v


def scipp_dtype_to_python(value: Variable):
    from scipp import DType
    dtype = value.dtype
    if dtype == DType.rotation3 or dtype == DType.vector3:
        # these are always floats
        return [float(v) for v in value.value]
    if dtype == DType.float64 or dtype == DType.float32:
        return float(value.value)
    elif dtype == DType.int64 or dtype == DType.int32:
        return int(value.value)
    return value.value


def serialize_scipp(value: Variable):
    """Convert scipp Variable objects to serializable dicts

    Only scalar and array objects used in this module are supported.
    Extending to new types of objects could be achieved.
    """
    from .spatial import __is_vector__, __is_quaternion__
    rep = {}
    if is_scalar(value):
        rep['type'] = 'scipp_scalar'
        rep['value'] = scipp_dtype_to_python(value)
    else:
        rep['type'] = 'scipp_array'
        rep['values'] = [scipp_dtype_to_python(v) for v in value]
        rep['dims'] = value.dims

    rep['dtype'] = str(value.dtype)
    rep['unit'] = str(value.unit)
    return rep


def serialize_mccode(value: InstrumentParameter | Expr):
    if isinstance(value, InstrumentParameter):
        print(f'Requested to make parameter {value} serializable')
        return {'type': 'mccode_instrument_parameter', 'value': str(value)}
    elif isinstance(value, Expr):
        return {'type': 'mccode_expr', 'value': value.value, 'scalar': value.is_scalar, 'dtype': str(value.data_type)}


def dprint(d: dict, n: int = 0):
    """Pretty print a dictionary with indent-depth indicating nesting"""
    pre = ' '*n
    for k, v in d.items():
        if isinstance(v, dict):
            print(f'{pre}{k}:')
            dprint(v, n + 1)
        else:
            print(f'{pre}{k}: {v}')


def compare(a, b, depth: str = None):
    """Compare two dictionaries, list, tuples, or numeric values for equivalency

    Nesting of dictionaries, lists, and tuples supported.
    """
    if depth is None:
        depth = ''
    if not isinstance(a, type(b)):
        return False
    if isinstance(a, dict):
        return compare_dict(a, b, depth=depth)
    if isinstance(a, (list, tuple)):
        if len(a) != len(b):
            return False
        return all(compare(x, y, depth=depth) for x, y in zip(a, b))
    return compare_one(a, b, depth=depth)


def compare_dict(dict1, dict2, depth: str):
    for kappa in dict2:
        if kappa not in dict1:
            return False
    for k, v in dict1.items():
        if k not in dict2:
            return False
        nu = dict2[k]
        return compare(v, nu, depth=f'{depth}/k')


def compare_one(a, b, depth: str):
    """Compare two things which are not dicts, lists, or tuples and are the same type"""
    from numpy import allclose
    if isinstance(a, str):
        return a == b
    return allclose(a, b)


def calibration_input(*args, **kwargs):
    """Allow specifying the calibration parameters as a single positional dict or
    any number of keyword arguments, with the latter taking precedence."""
    if len(args) == 1 and isinstance(args[0], dict):
        params = args[0].copy()
    elif len(args):
        raise ValueError('only one (dict) positional argument or keyword arguments are allowed')
    else:
        params = {}
    params.update(kwargs)
    return params


def calibration(func):
    """A decorator for, e.g. a `@classmethod` or @static decorated Class method which
    is the from_calibration function for the class to handle input collation"""
    def wrapper(*args, **kwargs):
        if len(args) and isinstance(args[0], type):
            # this is the @classmethod form, where args[0] _is_ the class
            return func(args[0], calibration_input(*args[1:], **kwargs))
        return func(calibration_input(*args, **kwargs))
    return wrapper



from scipp import scalar, Variable

mm = scalar(1.0, unit='mm').to(unit='m')
bandwidth_monitor = 15 * mm
attenuator_1 = 5 * mm
attenuator_2 = 10 * mm
attenuator_3 = 3 * mm

def guide_table():
    from .guide_tools import parse_guide_table

    after_bwc_device_gap = 63.5 * mm

    gap_pre_bandwidth_monitor = 2 * mm  # FIXME this must be verified
    gap_bandwidth_monitor_attenuator_1 = 2 * mm
    gap_attenuator_1_attenuator_2 = 2 * mm
    gap_attenuator_2_attenuator_3 = 2 * mm
    gap_post_attenuator_3 = after_bwc_device_gap - (
        gap_pre_bandwidth_monitor + bandwidth_monitor
        + gap_bandwidth_monitor_attenuator_1 + attenuator_1
        + gap_attenuator_1_attenuator_2 + attenuator_2
        + gap_attenuator_2_attenuator_3 + attenuator_3
    )


    swissneutronics_37835_straight_guide_section_table = (
        ('window', 0.5 * mm),
        ('window gap', 2.5 * mm),
        ('boral mask', 3.0 * mm),
        ('45776- Unit-29 straight', 1945.1 * mm),
        ('unit gap', 0.5 * mm),
        ('45777- Unit-30 straight', 1945.1 * mm),
        ('bellow gap', 4.0 * mm),
        ('45778- Unit-31 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45779- Unit-32 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45780- Unit-33 straight', 1781.89 * mm),
        ('unit gap', 0.5 * mm),
        ('45781- Unit-34 straight', 1781.89 * mm),
        ('bellow gap', 4.0 * mm),
        ('45782- Unit-35 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45783- Unit-36 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45784- Unit-37 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45785- Unit-38 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45786- Unit-39 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45787- Unit-40 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45788- Unit-41 straight', 1936.88 * mm),
        ('unit gap', 0.5 * mm),
        ('45789- Unit-42 straight', 1936.88 * mm),
        ('unit gap', 0.5 * mm),
        ('45790- Unit-43 straight', 1171.2 * mm),

        # ('device gap', 160.0 * mm),  # old idea to have monitor in vacuum

        ('unit gap', 24 * mm), # end of Unit-24 to first BWC disc
        ('device gap', 11 * mm), # width of first BWC disc
        ('unit gap', 24 * mm), # estimated gap from first disc to second disc
        ('device gap', 11 * mm), # second BWC disc width
        ('window gap', 24 * mm),
        ('window', 0.5 * mm),
        # ('device gap', 63.5 * mm), # location of bandwidth monitor then attenuators
        ('gap', gap_pre_bandwidth_monitor),
        ('device gap', bandwidth_monitor),
        ('gap', gap_bandwidth_monitor_attenuator_1),
        ('device gap', attenuator_1),  # 5 mm Plexiglass
        ('gap', gap_attenuator_1_attenuator_2),
        ('device gap', attenuator_2), # 10 mm Plexiglass
        ('gap', gap_attenuator_2_attenuator_3),
        ('device gap', attenuator_3), # 3 mm B4C with 68 2 mm diameter holes
        ('gap', gap_post_attenuator_3),

        ('window', 0.5 * mm),
        ('window gap', 2.5 * mm),

        ('boral mask', 3.0 * mm),
        ('45791 Unit-44 straight', 1406.4 * mm),
        ('unit gap', 0.5 * mm),
        ('45792- Unit-45 straight', 1406.4 * mm),
        ('bellow gap', 4.0 * mm),
        ('45793- Unit-46 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45794- Unit-47 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45795- Unit-48 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45796- Unit-49 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45797- Unit-50 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45798- Unit-51 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45799- Unit-52 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45800- Unit-53 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45801- Unit-54 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45802- Unit-55 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45803- Unit-56 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45804- Unit-57 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45805- Unit-58 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45806- Unit-59 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45807- Unit-60 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45808- Unit-61 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45809- Unit-62 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45810- Unit-63 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45811- Unit-64 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45812- Unit-65 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45813- Unit-66 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45814 Unit-67 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45815- Unit-68 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45816- Unit-69 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45817- Unit-70 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45818- Unit-71 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45819- Unit-72 straight', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45820- Unit-73 straight', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45821- Unit-74 straight', 1000.0 * mm),
        ('bellow gap', 4.0 * mm),
        ('45822- Unit-75 straight', 1311.52 * mm),
        ('unit gap', 0.5 * mm),
    )
    return parse_guide_table(swissneutronics_37835_straight_guide_section_table)


def unit_dict(ref_p, ref_r, number, length):
    # The straight guide is made up of entirely equivalent guide units
    from .guide_tools import straight_unit_dict
    params = {
        'width': 60 * mm,
        'height': 90 * mm,
        'left': 1.5,
        'right': 1.5,
        'top': 1.5,
        'bottom': 1.5,
        'length': length,
    }
    return straight_unit_dict(ref_p, ref_r, params)


def straight_guide_parameters(guide_pos, guide_rot, chopper_height) -> tuple[dict, Variable, Variable]:
    from scipp import array, vector
    from .guide_tools import guide_partial_dict, device_partial_dict, entering_partial_dict
    table = guide_table()
    beam = {'width': 60 * mm, 'height': 90 * mm}
    window = {k: v + 20 * mm for k, v in beam.items()}

    # finish the gap left for the shutter
    p, ref_p, ref_r = entering_partial_dict(guide_pos, guide_rot, None, table, window)

    # straight guide from shutter to bandwidth chopper
    d, ref_p, ref_r = guide_partial_dict(ref_p, ref_r, table, 29, 43, unit_dict)
    p.update(d)

    radius = 350 * mm
    offset = -(radius - chopper_height / 2) * vector([0, 1.0, 0])
    chopper = {
        'radius': 350 * mm,
        'height': chopper_height,
        'angle': scalar(161.0, unit='deg'),
        'frequency': scalar(14.0, unit='Hz'),
        'phase': scalar(0., unit='deg'),
        'offset': offset,
    }

    plexiglass = 'AcrylicGlass_C5O2H8'
    b4c = 'B4C_sg166_BoronCarbide' # TODO check whether this should be *enriched* 10B, https://github.com/mctools/ncrystal/discussions/286
    attenuator = {'width': 88 * mm, 'height': 105 * mm, 'composition': plexiglass}

    devices = (
        ('bandwidth_chopper_1', {**chopper}),
        ('bandwidth_chopper_2', {**chopper}),
        ('bandwidth_monitor', {'thickness': bandwidth_monitor, 'sample_rate': scalar(42., unit='kHz'), **beam}),
        ('attenuator_1', {'length': attenuator_1, **attenuator}),
        ('attenuator_2', {'length': attenuator_2, **attenuator}),
        ('attenuator_3', {'length': attenuator_3, **attenuator, 'composition': b4c}),
    )
    d, ref_p, ref_r = device_partial_dict(ref_p, ref_r, devices, table, 43, 44, window)
    p.update(d)

    # straight guide from after monitor to the start of the closing section
    d, ref_p, ref_r = guide_partial_dict(ref_p, ref_r, table, 44, 76, unit_dict)
    p.update(d)

    return p, ref_p, ref_r


from __future__ import annotations

from scipp import Variable
from enum import Enum


class Desc(Enum):
    gap = 1
    device_gap = 2
    window_gap = 3
    unit_gap = 4
    bellow_gap = 5
    boral_mask = 6


class Type(Enum):
    guide = 1
    gap = 2
    window = 3
    mask = 4


def parse_guide_table(table: tuple[tuple[str, Variable],...]):
    import re
#    r = re.compile(r'(?P<guide>(?P<number>[0-9]+)-\sUnit-(?P<id>[0-9]+)\s(?P<name>[a-zA-Z0-9]+))|(?P<gap>[a-z\s]*gap)')
    # Guides are normally '{number}- Unit-{id} {name}' but in at least one case the
    # hyphen after '{number}' is missing, so make it optional.
    r_guide = re.compile(r'(?P<number>[0-9]+)-?\sUnit-(?P<id>[0-9]+)\s(?P<name>[a-zA-Z0-9\s]+)')
    r_gap = re.compile(r'[a-z\s]*gap')
    r_mask = re.compile(r'[a-z\s]*mask')

    def lns(s):
        return s.lower().replace(' ', '_')

    def label_to_key(label):
        if (m := r_guide.match(label)) is not None:
            return Type.guide, int(m.group('number')), int(m.group('id')), lns(m.group('name'))
        if (m := r_gap.match(label)) is not None:
            return Type.gap, Desc[lns(m.group(0))]
        if (m := r_mask.match(label)) is not None:
            return Type.mask, Desc[lns(m.group(0))]
        # fallback on the specified label being a known type
        try:
            return (Type[label], )
        except ValueError:
            raise RuntimeError('Could not parse label "{}"'.format(label))

    return tuple(label_to_key(label) + (length,) for label, length in table)


def straight_unit_dict(ref_p, ref_r, params):
    """
    Construct the dict representation of a guide which is not rotated relative to
    its reference point and orientation, and for which the other guide parameters are
    known. Update the returned reference point to be at the exit of this guide unit.

    :param ref_p:
        The scipp.vector representing the starting point of this guides in the global
        coordinate system
    :param ref_r:
        The scipp.rotation_3 Quaternion representing the starting orientation of
        this guide in the global coordinate system
    :param params:
        A dictionary of parameters representing the guide, with the exception of its
        position or orientation.
    :return:
    """
    from scipp import vector
    from ..spatial import at_relative
    d = {
        'position': ref_p,
        'orientation': ref_r,
    }
    d.update(params)
    # move the reference point to the end of this guide
    if 'length' not in d:
        raise ValueError('Guide length required to move reference point')
    ref_p = at_relative(ref_p, ref_r, d['length'] * vector([0, 0, 1.]))
    return d, ref_p, ref_r


def guide_partial_dict(ref_p, ref_r, table, min_unit, max_unit, unit_function):
    """

    :param ref_p:
        The scipp.vector representing the starting point of this section of guides
        in the global coordinate system
    :param ref_r:
        The scipp.rotation_3 Quaternion representing the starting orientation of
        this section of guides in the global coordinate system
    :param table:
        A parsed table of guide information
    :param min_unit:
        The lower-bound guide unit number to include in this section of guides
    :param max_unit:
        The upper-bound of guide unit number to include in this section of guides
    :param unit_function:
        A function which takes a position, quaternion, and unique guide number and
        returns the dictionary representing that guide, plus the position
        and orientation of the guide *exit*, as a 3-tuple
    :return:
        The collective dictionary representing all guide elements in this guide
        segment, the final guide-exit position, and the final guide-exit orientation.
    """
    from scipp import vector
    from ..spatial import at_relative
    in_section = False
    d = {}
    for entry in table:
        t = entry[0]
        if Type.guide == t:
            in_section = min_unit <= entry[2] <= max_unit
        if in_section:
            if Type.guide == t:
                number, section, name = entry[1:-1]
                d[f'unit_{section}_{name}'], ref_p, ref_r = unit_function(ref_p, ref_r, number, length=entry[-1])
            elif Type.window == entry[0]:
                raise ValueError(f'No windows allowed in continuous section {min_unit}-{max_unit}!')
            else:
                ref_p = at_relative(ref_p, ref_r, entry[-1] * vector([0, 0, 1.]))
        if Type.guide == t:
            in_section = min_unit <= entry[2] < max_unit

    return d, ref_p, ref_r


def device_partial_dict(ref_p: Variable, ref_r: Variable, device: tuple[tuple[str, dict], ...] | None,
                        table, pre_unit, post_unit, al_consts: dict):
    """
    Insert one or more device into an equal number of 'device gaps' in a straight guide

    No orientation change is allowed between devices inserted into subsequent gaps.

    :param ref_p:
        The scipp.vector representing the starting point of this section of non-guides
        in the global coordinate system
    :param ref_r:
        The scipp.rotation_3 Quaternion representing the starting orientation of
        this section of non-guides in the global coordinate system
    :param device:
        A tuple of tuples, each containing the device name and partial dictionary
        representing the device information to be placed into a gap in the guide.
        The order of devices should match the order of device gaps.
    :param table:
        A parsed table of guide information
    :param pre_unit:
        The last guide unit before the section containing the device gap
    :param post_unit:
        The first guide unit after the section containing the device gap
    :param al_consts:
        Any parameters for aluminum windows, exiting the first unit and entering
        the second. Notably the width and height of the windows should be provided,
        their thickness should come from the table but could be overruled here.
    :return:
    """
    from scipp import vector
    from ..spatial import at_relative

    d, section, pre_ok, post_ok, is_exit = {}, 0, False, False, True

    index = 0
    if device is None:
        device = (('LOGICAL_ERROR_IN_' + __file__, {}),)

    for entry in table:
        t = entry[0]
        if Type.guide == t:
            section = entry[2]
            pre_ok = section >= pre_unit
            post_ok = section < post_unit
        if pre_ok and post_ok and Type.guide != t:
            dist = entry[-1]
            if Type.window == t:
                n, s = (section, 'exit') if is_exit else (section + 1, 'entry')
                d[f'unit_{n}_{s}_window'] = {
                    'position': ref_p,
                    'orientation': ref_r,
                    'length': dist,
                    'composition': 'Al_sg225',
                    **al_consts,
                }
                is_exit = False
            elif Type.gap == t and Desc.device_gap == entry[1]:
                dev_name, dev_dict = device[index]
                index += 1

                half_p = at_relative(ref_p, ref_r, dist/2 * vector([0, 0, 1.]))
                if 'offset' in dev_dict:
                    half_p -= dev_dict['offset']
                d[dev_name] = {'position': half_p, 'orientation': ref_r}
                d[dev_name].update(dev_dict)

            ref_p = at_relative(ref_p, ref_r, dist * vector([0, 0, 1.]))

    return d, ref_p, ref_r


def entering_partial_dict(ref_p: Variable, ref_r: Variable,
                          device: tuple[tuple[str, dict], ...] | None,
                          table, al_consts: dict):
    """
    Insert one or more device into an equal number of 'device gaps' in a straight guide
    which all must come before the first guide section in the provided table

    No orientation change is allowed between devices inserted into subsequent gaps.

    :param ref_p:
        The scipp.vector representing the starting point of this section of non-guides
        in the global coordinate system
    :param ref_r:
        The scipp.rotation_3 Quaternion representing the starting orientation of
        this section of non-guides in the global coordinate system
    :param device:
        A tuple of tuples, each containing the device name and partial dictionary
        representing the device information to be placed into a gap in the guide.
        The order of devices should match the order of device gaps.
    :param table:
        A parsed table of guide information
    :param al_consts:
        Any parameters for aluminum windows, exiting the first unit and entering
        the second. Notably the width and height of the windows should be provided,
        their thickness should come from the table but could be overruled here.
    :return:
    """
    from scipp import vector
    from ..spatial import at_relative

    d, section, pre_ok, post_ok, is_exit = {}, 0, False, False, True

    index = 0
    if device is None:
        device = (('LOGICAL_ERROR_IN_' + __file__, {}),)

    # get the section number which we will enter into, for the (presumably) present window
    for entry in table:
        if Type.guide == entry[0]:
            section = entry[2]
            break

    for entry in table:
        t = entry[0]
        if Type.guide == t:
            break
        else:
            dist = entry[-1]
            if Type.window == t:
                d[f'unit_{section}_entry_window'] = {
                    'position': ref_p,
                    'orientation': ref_r,
                    'length': dist,
                    'composition': 'Al_sg225',
                    **al_consts,
                }
                is_exit = False
            elif Type.gap == t and Desc.device_gap == entry[1]:
                dev_name, dev_dict = device[index]
                index += 1

                half_p = at_relative(ref_p, ref_r, dist/2 * vector([0, 0, 1.]))
                if 'offset' in dev_dict:
                    half_p -= dev_dict['offset']
                d[dev_name] = {'position': half_p, 'orientation': ref_r}
                d[dev_name].update(dev_dict)

            ref_p = at_relative(ref_p, ref_r, dist * vector([0, 0, 1.]))

    return d, ref_p, ref_r



def exiting_partial_dict(ref_p: Variable, ref_r: Variable, table, al_consts: dict):
    """
    Insert an aluminum window in a straight guide following the last guide unit

    :param ref_p:
        The scipp.vector representing the starting point of this section of non-guides
        in the global coordinate system
    :param ref_r:
        The scipp.rotation_3 Quaternion representing the starting orientation of
        this section of non-guides in the global coordinate system
    :param table:
        A parsed table of guide information
    :param al_consts:
        Any parameters for aluminum windows, exiting the first unit and entering
        the second. Notably the width and height of the windows should be provided,
        their thickness should come from the table but could be overruled here.
    :return:
    """
    from scipp import vector
    from ..spatial import at_relative

    d, section, pre_ok, post_ok, is_exit = {}, 0, False, False, True

    index = 0

    # get the last section number which we will leave
    for i, entry in enumerate(table):
        if Type.guide == entry[0]:
            section = entry[2]
            index = i

    for entry in table[index:]:
        t = entry[0]
        # No need to worry about the units before the last one, since we use `index`
        if Type.guide != t:
            dist = entry[-1]
            if Type.window == t:
                d[f'unit_{section}_exit_window'] = {
                    'position': ref_p,
                    'orientation': ref_r,
                    'length': dist,
                    'composition': 'Al_sg225',
                    **al_consts,
                }
            ref_p = at_relative(ref_p, ref_r, dist * vector([0, 0, 1.]))

    return d, ref_p, ref_r



from scipp import Variable
from .component import Component


class Collimator(Component):
    """A device which limits the horizontal divergence and/or scattering kernel size"""
    width: Variable
    height: Variable
    length: Variable
    blades: int  # number of infinitely thin blades that fit within width


class SollerCollimator(Collimator):
    """Collimator with linear width, height, and length"""
    pass


class RadialCollimator(Collimator):
    """Collimator with angular width, linear height, and radial length"""
    radius: Variable  # the inner radius of the collimator blades




from scipp import Variable


def tube_xy_displacement_to_quaternion(length: Variable, displacement: Variable):
    from scipp import vector, scalar, any, sqrt, allclose
    from ..spatial import vector_to_vector_quaternion
    com_to_end = length * vector([0, 0, 0.5]) + displacement
    l2 = length * length
    x2 = displacement.fields.x * displacement.fields.x
    y2 = displacement.fields.y * displacement.fields.y

    com_to_end.fields.z = sqrt(0.25 * l2 - x2 - y2)

    z2 = displacement.fields.z * displacement.fields.z
    if any(z2 > scalar(0, unit=z2.unit)) and not allclose(com_to_end.fields.z, 0.5 * length - displacement.fields.z):
        raise RuntimeError("Provided tube-end displacement vector(s) contain wrong z-component value(s)")

    # The tube *should* point along z, but we were told it is displaced in x and y;
    # return the orienting Quaternion that takes (001) to the actual orientation
    quaternion = vector_to_vector_quaternion(vector([0, 0, 1]), com_to_end)
    return quaternion


def known_pack_params():
    from scipp import array, scalar, vector, vectors
    from numpy import arange
    known = dict()
    known['sample_detector_distance'] = scalar(3.5, unit='m')
    known['detector_length'] = scalar(3.5, unit='m')
    known['detector_radius'] = scalar(25.4/2, unit='mm')
    known['detector_orient'] = vector([0, 0, 0], unit='mm')
    known['sample'] = vector([0, 0, 0], unit='m')
    known['resistance'] = scalar(380, unit='Ohm')
    known['resistivity'] = scalar(140, unit='Ohm/in').to(unit='Ohm/m')
    # each pack is 32 tubes with 0.43 degrees between tubes
    known['tube_angles'] = array(values=0.43*(arange(32)-16+0.5), unit='degree', dims=['tube'])
    # 5 degree minimum scattering angle, centers 33*0.43 degrees apart such that the highest angle is 180-3
    known['pack_angles'] = array(values=5 + 16*0.43 + (33 * 0.43) * arange(7), unit='degree', dims=['pack'])
    return known



from dataclasses import dataclass


@dataclass
class BIFROST:
    from ..components import IndirectSecondary
    # primary: BandwidthPrimary   # A minimal form of the primary spectrometer necessary to transform events
    secondary: IndirectSecondary  # A minimal form of the secondary spectrometer necessary to transform events

    @staticmethod
    def from_calibration(**params):
        from scipp import vector
        from .tank import Tank
        # primary = ...
        params['sample'] = params.get('sample', vector([0, 0, 0.], unit='m'))
        tank = Tank.from_calibration(**params)

        secondary = tank.to_secondary(sample=params['sample'])
        return BIFROST(secondary)



# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT


def get_mccode_registries():
    from mccode_antlr.reader import GitHubRegistry

    registries = ['mcstas-chopper-lib', 'mcstas-transformer', 'mcstas-detector-tubes',
                  'mcstas-epics-link', 'mcstas-frame-tof-monitor', 'mccode-mcpl-filter',
                  'mcstas-monochromator-rowland', 'mcstas-slit-radial']
    registries = [GitHubRegistry(
        name,
        url=f'https://github.com/mcdotstar/{name}',
        filename='pooch-registry.txt',
        version='main'
    ) for name in registries]

    return registries



def test_bifrost_whole():
    from niess.bifrost.parameters import primary_parameters, tank_parameters
    from niess.bifrost import Tank, Primary

    primary = Primary.from_calibration(primary_parameters())
    tank = Tank.from_calibration(tank_parameters())


def test_bifrost_mccode():
    from niess.bifrost.parameters import primary_parameters, tank_parameters
    from niess.bifrost import Tank, Primary
    from mccode_antlr import Flavor
    from mccode_antlr.assembler import Assembler

    bifrost = Assembler('bifrost', registries=get_mccode_registries(), flavor=Flavor.MCSTAS)

    primary = Primary.from_calibration(primary_parameters())
    primary.to_mccode(bifrost)

    # TODO insert pre- and post-sample things here
    #      e.g., the split_at location at the end of the guide
    #      any filters, e.g., a hits-the-sample MCPL filter, or a Be-transmission filter
    #      the radial collimator between sample and tank, etc.

    tank = Tank.from_calibration(tank_parameters())
    tank.to_mccode(bifrost, 'sample_origin')

    # TODO add checks that conversion from the intermediate representation works?



from .secondary import DirectSecondary, IndirectSecondary
from .crystals import IdealCrystal, Crystal
from .detectors import Wire, DiscreteWire, DiscreteTube, He3Tube
from .aperture import Aperture, Jaw, Slit
from .chopper import Chopper, DiscChopper, FermiChopper
from .collimator import Collimator, SollerCollimator, RadialCollimator
from .component import Component
from .filter import Attenuator, Filter, OrderedFilter, make_aluminum
from .guide import EllipticGuide, TaperedGuide, StraightGuide, Guide, StraightGuides, TaperedGuides
from .moderator import Moderator
from .monitors import FissionChamber, He3Monitor, BeamCurrentMonitor, GEM2D
from .source import ESSource
from .section import Section

__all__ = [
    DirectSecondary,
    IndirectSecondary,
    IdealCrystal,
    Crystal,
    Wire,
    DiscreteWire,
    DiscreteTube,
    He3Tube,
    Aperture,
    Jaw,
    Slit,
    Chopper,
    DiscChopper,
    FermiChopper,
    Collimator,
    SollerCollimator,
    RadialCollimator,
    Component,
    Attenuator,
    Filter,
    OrderedFilter,
    make_aluminum,
    Guide,
    EllipticGuide,
    TaperedGuide,
    StraightGuide,
    StraightGuides,
    TaperedGuides,
    Moderator,
    FissionChamber,
    He3Monitor,
    BeamCurrentMonitor,
    GEM2D,
    ESSource,
    Section,
]


from __future__ import annotations

from niess.components.component import Base
from typing import ClassVar, Type

class Arm(Base):
    from networkx import DiGraph
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.instr import Instance
    from .analyzer import Analyzer
    from .triplet import Triplet
    from scipp import Variable

    analyzer: Analyzer
    detector: Triplet

    __struct_field_types__: ClassVar[dict[str, Type]] = {'analyzer': Analyzer, 'detector': Triplet}

    @classmethod
    def from_dict(cls, data):
        from .analyzer import Analyzer
        from .triplet import Triplet
        analyzer = data['analyzer']
        detector = data['detector']
        if not isinstance(analyzer, Analyzer):
            analyzer = Analyzer.from_dict(analyzer)
        if not isinstance(detector, Triplet):
            detector = Triplet.from_dict(detector)
        return cls(analyzer, detector)

    @staticmethod
    def from_calibration(a_position, tau, d_position, d_length, **params):
        from .analyzer import Analyzer
        from .triplet import Triplet

        analyzer_orient = params.get('analyzer_orient', None)
        detector_orient = params.get('detector_orient', None)
        # the analyzer focuses on the center tube of the triplet
        a_focus = d_position['tube', 1] if 'tube' in d_position.dims else d_position
        analyzer = Analyzer.from_calibration(a_position, a_focus, tau, **params, orient=analyzer_orient)
        detector = Triplet.from_calibration(d_position, d_length, **params, orient=detector_orient)
        return Arm(analyzer, detector)

    def triangulate_detector(self, unit=None):
        return self.detector.triangulate(unit=unit)

    def triangulate_analyzer(self, unit=None):
        return self.analyzer.triangulate(unit=unit)

    def triangulate(self, unit=None):
        from ..spatial import combine_triangulations
        return combine_triangulations([self.triangulate_analyzer(unit=unit), self.triangulate_detector(unit=unit)])

    def mcstas_parameters(self, sample: Variable):
        from numpy import stack, hstack
        from scipp import sqrt, dot, cross, vector, acos
        from ..spatial import is_scipp_vector, perpendicular_directions
        is_scipp_vector(sample, 'sample')

        # TODO find sample-analyzer and analyzer-detector distances, move positions into appropriate frames
        # analyzer_position -> [0, 0, sample-analyzer-distance]
        # detector_position -> [[dx0, dy0, dz0], [dx1, dy1, analyzer-detector-distance], [dx2, dy2, dz2]]
        # end_position -> z along analyzer-detector vector 'Arm' (in McStas local coordinate frame)

        sa = self.analyzer.central_blade.position - sample
        ad = (self.detector.tubes[1].at + self.detector.tubes[1].to)/2 - self.analyzer.central_blade.position
        distances = [sqrt(dot(x, x)).to(unit='m').value for x in (sa, ad)]
        # the coordinate system here has 'local' x along the beam, and z vertical
        # the McStas local cooridnate system always has z along the beam and y defines the local scattering plane normal
        # for BIFROST's analyzers, the two coordinate systems have parallel (or maybe antiparallel) y directions

        za = sa / sqrt(dot(sa, sa))
        zd = ad / sqrt(dot(ad, ad))
        yd = cross(za, zd)
        yd /= sqrt(dot(yd, yd))
        xd = cross(yd, zd)

        two_theta = acos(dot(za, zd))

        tube_com = self.detector.tube_com() - self.analyzer.central_blade.position
        tube_end = self.detector.tube_end()

        x, y, z = [vector(q) for q in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]]

        # this could be simplified if we built the column matrix (xd, yd, zd)
        tube_com_x, tube_com_y, tube_com_z = [dot(tube_com, d) * i for d, i in zip((xd, yd, zd), (x, y, z))]
        tube_com_d = tube_com_x + tube_com_y + tube_com_z
        tube_end_x, tube_end_y, tube_end_z = [dot(tube_end, d) * i for d, i in zip((xd, yd, zd), (x, y, z))]
        tube_end_d = tube_end_x + tube_end_y + tube_end_z
        # shift the COM relative to the expected detector position
        tube_com_d.fields.z -= sqrt(dot(ad, ad))

        # this is not good. Can we verify which axis is the coordinate axis and which is the tube axis?
        d = stack((tube_com_d.to(unit='m').values, tube_end_d.to(unit='m').values), axis=1)

        hc, vc = self.analyzer.coverage(sample)
        a = hstack((self.analyzer.count, self.analyzer.central_blade.shape.to(unit='m').value, [hc.value, vc.value]))

        return {'distances': distances, 'analyzer': a, 'detector': d, 'two_theta': two_theta.value}

    def rtp_parameters(self, sample: Variable):
        from scipp import concat, cross, dot, sqrt
        sa = self.analyzer.central_blade.position - sample
        ad = (self.detector.tubes[1].at + self.detector.tubes[1].to)/2 - self.analyzer.central_blade.position

        out_of_plane = cross(ad, sa)
        x, y, angle = self.analyzer.rtp_parameters(sample, out_of_plane)
        return sqrt(dot(sa, sa)), sqrt(dot(ad, ad)), x, y, angle

    def sample_space_angle(self, sample: Variable):
        return self.analyzer.sample_space_angle(sample)

    def coverage(self, sample: Variable, unit=None):
        # The arm coverage is defined vertically by the analyzer, but the mean
        # horizontal divergence is defined by the active *detector* length
        unit = unit or 'radian'
        ana_hor, ana_ver = self.analyzer.coverage(sample, unit=unit)
        det_hor = self.detector.horizontal_coverage(sample, self.analyzer.central_blade.position, unit=unit)
        if det_hor > ana_hor:
            print(f'Detector under-illuminated: the detector width {det_hor} should be less than the analyzer width {ana_hor}')
        return det_hor, ana_ver

    def to_mccode(self, assembler: Assembler, ref: Instance, name: str,
                  analyzer_when: str = None, analyzer_extend: str = None,
                  detector_when: str = None, detector_extend: str = None, **kwargs):
        from scipp import concat, all, isclose, vector, dot, sqrt, atan2
        # For each channel we need to define the local coordinate system, relative to the provided sample
        origin = vector([0, 0, 0], unit='m')

        sa_vec = self.analyzer.central_blade.position
        ad_vec = (self.detector.tubes[1].at + self.detector.tubes[1].to) / 2 - sa_vec

        sample_analyzer_d = sqrt(dot(sa_vec, sa_vec)).to(unit='m')
        analyzer_detector_distance = sqrt(dot(ad_vec, ad_vec)).to(unit='m')

        x = dot(ad_vec, sa_vec / sample_analyzer_d)
        y = dot(ad_vec, vector([0, 0, 1]))
        two_theta = atan2(y=y, x=x).to(unit='degree').value
        theta = two_theta / 2

        point = f'{name}_analyzer_point'    # component name of the location of the analyzer
        mono = f'{name}_monochromator'      # component name of the analyzer itself
        orient = f'{name}_detector_angle'   # component name of the oriented arm pointing at the detector
        triplet = f'{name}_triplet'         # component name of the detector itself

        # Move to the center of the analyzer & reorient for monochromator scattering in vertical plane
        arm = assembler.component(point, "Arm", at=((0, 0, sample_analyzer_d.value), ref), rotate=((0, 0, 90), ref))
        if analyzer_when is not None:
            arm.WHEN(analyzer_when)
        # Insert the analyzer rotated by theta (origin is used for calculating coverage angles)
        self.analyzer.to_mccode(assembler, source=ref.name, relative=point, sink=triplet, theta=theta, name=mono,
                                when=analyzer_when, extend=analyzer_extend, origin=origin)
        # Change the coordinate system by theta -- total scattering angle is then 2theta
        det_angle = assembler.component(orient, "Arm", at=((0, 0, 0), mono), rotate=((0, theta, 0), mono))
        det_angle.WHEN(detector_when)
        # Insert the detector distance along that arm
        self.detector.to_mccode(assembler, relative=orient, distance=analyzer_detector_distance.value, name=triplet,
                                when=detector_when, extend=detector_extend,
                                component=kwargs.get('detector_component', None),
                                parameters=kwargs.get('detector_parameters', None))

    def add_to_graph(self, upstream: str | None, name: str, graph: DiGraph):
        point = f'{name}_analyzer_point'  # component name of the location of the analyzer
        mono = f'{name}_monochromator'  # component name of the analyzer itself
        orient = f'{name}_detector_angle'  # component name of the oriented arm pointing at the detector
        triplet = f'{name}_triplet'  # component name of the detector itself

        graph.add_node(point)
        if upstream is not None:
            graph.add_edge(upstream, point)
        self.analyzer.add_to_graph(point, mono, graph)
        graph.add_node(orient)
        graph.add_edge(mono, orient)
        self.detector.add_to_graph(orient, triplet, graph)

        return [triplet]




import msgspec
from typing import Callable

class MarkdownRow(msgspec.Struct):
    contents: tuple[list[str], ...]
    lengths: tuple[int, ...]
    is_heading: bool = False

    def row_lengths(self):
        def max_or_zero(col):
            if col is None or len(col) == 0:
                return 0
            return max(len(c) for c in col)

        return [max_or_zero(col) for col in self.contents]

    def __post_init__(self):
        def correct(content):
            if isinstance(content, str):
                return content.split('\n')
            return content
        self.contents = tuple(correct(c) for c in self.contents)
        self.lengths = tuple(self.row_lengths())

    @classmethod
    def from_json(cls, data, headings: list[str], transforms: dict[str,Callable]):
        if isinstance(data, (str, bytes)):
            from json import loads
            data = loads(data)
        contents = [transforms[heading](data[heading]) for heading in headings]
        lengths = [0 for _ in headings]
        return cls(tuple(contents), tuple(lengths))

    def __str__(self):
        rep = ''
        height = 1 + max(len(content) for content in self.contents)
        # Top line (if heading)
        if self.is_heading:
            for length in self.lengths:
                rep += '+' + '-' * (length + 2)
            rep += '+\n'
        # Contents line(s)
        for line in range(height):
            for col, length in zip(self.contents, self.lengths):
                if len(col) > line:
                    rep += f'| {col[line]:{length}} '
                else:
                    rep += '|' + ' ' * (length + 2)
            rep += '|\n'
        # Bottom line
        for length in self.lengths:
            rep += '+' + ('=' if self.is_heading else '-') * (length + 2)
        rep += '+\n'
        return rep

class MarkdownTable(msgspec.Struct):
    title: str
    rows: list[MarkdownRow]

    def __post_init__(self):
        self.rows[0].is_heading = True
        lengths = self.rows[0].lengths
        for row in self.rows[1:]:
            lengths = [max(a, b) for a, b in zip(lengths, row.lengths)]
        for row in self.rows:
            row.lengths = tuple(lengths)

    def __str__(self):
        return ''.join(str(row) for row in self.rows)


def to_str(value):
    if hasattr(value, '__len__') and isinstance(value[0], float):
        return '(' + ','.join(f'{x:0.3f}' for x in value) + ')'
    return str(value)

def scipp_variable_transform(x):
    if not isinstance(x, dict) or 'obj' not in x:
        raise ValueError(f'{x} is not a scipp variable?')
    o = x['obj']
    value = o.get('value', o.get('values', None))
    unit = o.get('unit')
    dims = o.get('dims')
    ret = to_str(value)
    if unit is not None and unit != 'dimensionless':
        ret += f' {unit}'
    if dims is not None:
        ret += f' {dims}'
    return ret


def parameter_one(d, depth):
    if isinstance(d, dict) and 'name' in d and 'scipp.Variable' == d['name']:
        return scipp_variable_transform(d)
    elif isinstance(d, dict):
        return '\n' + '\n'.join(parameter_dict(d, depth+1))
    return str(d)


def parameter_dict(d, depth=0):
    return ['  '*depth + f'- {k} = {parameter_one(v, depth)}' for k, v in d.items()]


def parameters_transform(parameters):
    return '\n'.join(parameter_dict(parameters))


def to_markdown(obj):
    from json import loads
    from .json import to_json
    data = loads(to_json(obj))

    if not 'name' in data or not 'obj' in data:
        raise TypeError('Missing "name" and "obj" fields')

    title = data['name']
    data = data['obj']

    # Data is now an (ordered) dictionary, with keys equal to component names
    headings = ['name', 'position', 'orientation', 'parameters']
    transformations = {
        'name': lambda x: f'`{x}`',
        'position': scipp_variable_transform,
        'orientation': scipp_variable_transform,
        'parameters': parameters_transform,
    }

    def one_entry(entry):
        name, pos, ori = entry.pop('name'), entry.pop('position'), entry.pop('orientation')
        entry = dict(name=name, position=pos, orientation=ori, parameters=entry)
        return MarkdownRow.from_json(entry, headings, transformations)

    def one_component(component):
        ret = []
        if 'segments' in component:
            for segment in component['segments']:
                ret.append(one_entry(segment))
        else:
            ret.append(one_entry(component))
        return ret

    rows = [r for value in data.values() for r in one_component(value)]
    row0 = MarkdownRow(headings, headings)
    table = MarkdownTable(title, [row0] + rows)
    return table


from typing import Any, Type
import msgspec
from scipp import Variable, DType


SCIPP_MODEL_ENCODE = {
    Variable: 'scipp.Variable'
}
SCIPP_MODEL_DECODE = {v: k for k, v in SCIPP_MODEL_ENCODE.items()}

def variable_to_dict(v: Variable) -> dict[str, Any]:
    d = {'unit': str(v.unit), 'dtype': str(v.dtype)}
    def tolist(x):
        if d['dtype'] == 'string' and v.size == 1:
            return x
        if hasattr(x, 'tolist'):
            return x.tolist()
        return x

    if v.size == 1:
        d['value'] = tolist(v.value)
        if v.variance is not None:
            d['variance'] = tolist(v.variance)
    else:
        d['values'] = tolist(v.values)
        d['dims'] = list(v.dims)
        if v.variances is not None:
            d['variances'] = tolist(v.variances)
    return d

def variable_json_schema():
    schema = {
        "$defs": {
            "nd_array": {
                "type": "array",
                "items": {"type": ["string", "integer", "number", {"$ref": "#/$defs/nd_array"}]},
                "minItems": 1
            }
        },
        'type': 'object',
        "title": "Scipp Variable",
        "description": "A scalar or array variable in scipp",
        "properties": {
            'unit': {"type": "string"},
            'dtype': {"type": "string"},
            "value": {"type": "number"},
            "variance": {"type": "number"},
            "values": {"type": {"$ref": "#/$defs/nd_array"}},
            "variances": {"type": {"$ref": "#/$defs/nd_array"}},
            "dims": {"type": "array", "items": {"type": "string"}},
        },
        "required": ["unit", "dtype", {"oneOf": ["value", "values"]}]
    }
    return schema


def dict_to_variable(d: dict) -> Variable:
    from scipp import array, scalar
    if isinstance(dtype := d['dtype'], str) and hasattr(DType, dtype):
        d['dtype'] = getattr(DType, dtype)
    if 'values' in d:
        return array(**d)
    else:
        return scalar(**d)


SCIPP_TO_DICT = {
    Variable: variable_to_dict,
}
DICT_TO_SCIPP = {
    'scipp.Variable': dict_to_variable,
}

SCIPP_JSON_SCHEMA = {Variable: variable_json_schema}


class ScippModel(msgspec.Struct):
    name: str
    obj: msgspec.Raw

    @classmethod
    def from_value(cls, obj: Any, encoder=None):
        if encoder is None:
            raise ValueError("An encoder must be provided")
        model_type = SCIPP_MODEL_ENCODE[type(obj)]
        if 'scipp.Variable' == model_type:
            obj = variable_to_dict(obj)
        return cls(model_type, msgspec.Raw(encoder.encode(obj)))


def to_scipp_model(scipp_model_encoder, obj):
    return ScippModel(scipp_model_encoder, obj)


def from_scipp_model(scipp_model_decoder, msg: ScippModel):
    if msg.name in SCIPP_MODEL_DECODE:
        return decode_scipp_model(scipp_model_decoder.decode(msg.obj))
    raise TypeError(f"{msg.name} is not a valid scipp model")

def decode_scipp_model(data: dict):
    return dict_to_variable(data)


"""The BIFROST primary spectrometer contains components which are only passed in series.

To make it easier to comprehend and maintain, the spectrometer is broken up into a
series of sections, each of which performs a specific function.

Each section contains a mixture of guide elements, choppers, monitors, attenuators,
windows, etc.; but they all come together into the primary spectrometer.
"""
from ..components import (
    Section, Component,
    Jaw, Slit, Filter, Attenuator, DiscChopper,
    EllipticGuide, StraightGuide, StraightGuides, TaperedGuides,
    FissionChamber, BeamCurrentMonitor, GEM2D, ESSource
)
from ..utilities import calibration

PRIMARY_TYPES = {
    'source': ESSource,
    # compressor section
    'nboa_entry_window': Filter,
    'nboa': EllipticGuide,  # Neutron Beam Optics Assembly
    'nboa_exit_window': Filter,
    'monolith_window': Filter,
    'bbg_entry_window': Filter,
    'bbg': EllipticGuide,  # Bridge Beam Guide
    'bbg_exit_window': Filter,
    'psc_housing_entry_window': Filter,
    'nose': EllipticGuide,
    # pulse shaping choppers (not in a named section ...)
    'pulse_shaping_chopper_1': DiscChopper,
    'pulse_shaping_chopper_2': DiscChopper,
    # curved guide section
    'unit_3_curved': StraightGuides,
    'psc_exit_window': Filter,
    'psc_monitor': FissionChamber,
    'curved_entrance_window': Filter,
    'unit_4_curved': StraightGuides,
    'unit_4_exit_window': Filter,
    'frame_overlap_chopper_1': DiscChopper,
    'unit_5_entry_window': Filter,
    'unit_5_curved': StraightGuides,
    'unit_6_curved': StraightGuides,
    'unit_7_curved': StraightGuides,
    'unit_8_curved': StraightGuides,
    'unit_8_exit_window': Filter,
    'frame_overlap_chopper_2': DiscChopper,
    'unit_9_entry_window': Filter,
    'unit_9_curved': StraightGuides,
    'unit_10_curved': StraightGuides,
    'unit_11_curved': StraightGuides,
    'unit_12_curved': StraightGuides,
    'unit_13_curved': StraightGuides,
    'unit_14_curved': StraightGuides,
    'unit_15_curved': StraightGuides,
    # expanding guide section
    'unit_16_bw_insert': EllipticGuide,  # Bunker wall insert, first part
    'unit_17_bw_insert': EllipticGuide,  # Bunker wall insert, second part
    'unit_17_exit_window': Filter,
    'overlap_monitor': BeamCurrentMonitor,
    'unit_18_entry_window': Filter,
    'unit_18_expanding': EllipticGuide,
    'unit_19_expanding': EllipticGuide,
    'unit_20_expanding': EllipticGuide,
    'unit_21_expanding': EllipticGuide,
    'unit_22_expanding': EllipticGuide,
    'unit_23_expanding': EllipticGuide,
    'unit_24_expanding': EllipticGuide,
    'unit_25_expanding': EllipticGuide,
    'unit_26_expanding': EllipticGuide,
    'unit_27_expanding': EllipticGuide,
    'unit_28_expanding': EllipticGuide,
    'unit_28_exit_window': Filter,
    # straight guide transport section
    'unit_29_entry_window': Filter,
    'unit_29_straight': StraightGuide,
    'unit_30_straight': StraightGuide,
    'unit_31_straight': StraightGuide,
    'unit_32_straight': StraightGuide,
    'unit_33_straight': StraightGuide,
    'unit_34_straight': StraightGuide,
    'unit_35_straight': StraightGuide,
    'unit_36_straight': StraightGuide,
    'unit_37_straight': StraightGuide,
    'unit_38_straight': StraightGuide,
    'unit_39_straight': StraightGuide,
    'unit_40_straight': StraightGuide,
    'unit_41_straight': StraightGuide,
    'unit_42_straight': StraightGuide,
    'unit_43_straight': StraightGuide,
    'bandwidth_chopper_1': DiscChopper,
    'bandwidth_chopper_2': DiscChopper,
    'unit_43_exit_window': Filter,
    'bandwidth_monitor': BeamCurrentMonitor,
    'attenuator_1': Attenuator,
    'attenuator_2': Attenuator,
    'attenuator_3': Attenuator,
    'unit_44_entry_window': Filter,
    'unit_44_straight': StraightGuide,
    'unit_45_straight': StraightGuide,
    'unit_46_straight': StraightGuide,
    'unit_47_straight': StraightGuide,
    'unit_48_straight': StraightGuide,
    'unit_49_straight': StraightGuide,
    'unit_50_straight': StraightGuide,
    'unit_51_straight': StraightGuide,
    'unit_52_straight': StraightGuide,
    'unit_53_straight': StraightGuide,
    'unit_54_straight': StraightGuide,
    'unit_55_straight': StraightGuide,
    'unit_56_straight': StraightGuide,
    'unit_57_straight': StraightGuide,
    'unit_58_straight': StraightGuide,
    'unit_59_straight': StraightGuide,
    'unit_60_straight': StraightGuide,
    'unit_61_straight': StraightGuide,
    'unit_62_straight': StraightGuide,
    'unit_63_straight': StraightGuide,
    'unit_64_straight': StraightGuide,
    'unit_65_straight': StraightGuide,
    'unit_66_straight': StraightGuide,
    'unit_67_straight': StraightGuide,
    'unit_68_straight': StraightGuide,
    'unit_69_straight': StraightGuide,
    'unit_70_straight': StraightGuide,
    'unit_71_straight': StraightGuide,
    'unit_72_straight': StraightGuide,
    'unit_73_straight': StraightGuide,
    'unit_74_straight': StraightGuide,
    'unit_75_straight': StraightGuide,
    # focusing section
    'unit_76_closing': EllipticGuide,
    'unit_77_closing': EllipticGuide,
    'unit_78_closing': EllipticGuide,
    'unit_79_closing': EllipticGuide,
    'unit_80_closing': EllipticGuide,
    'unit_81_closing': EllipticGuide,
    'unit_82_closing': EllipticGuide,
    'unit_83_closing': EllipticGuide,
    'unit_84_closing': EllipticGuide,
    'unit_85_closing': EllipticGuide,
    'jaw_3': Jaw,
    'unit_86_closing': EllipticGuide,
    'jaw_2': Jaw,
    'unit_87_closing': EllipticGuide,
    'jaw_1': Jaw,
    'unit_88_closing': EllipticGuide,
    'unit_88_exit_window': Filter,
    'mask': Slit,
    'normalization_monitor': GEM2D,
    'slit': Slit,
    'sample_origin': Component,
}

class Primary(Section):
    __struct_field_types__ = PRIMARY_TYPES  # an inconvenient hack

    source: ESSource

    # compressor section
    nboa_entry_window: Filter
    nboa: EllipticGuide  # Neutron Beam Optics Assembly
    nboa_exit_window: Filter
    monolith_window: Filter
    bbg_entry_window: Filter
    bbg: EllipticGuide  # Bridge Beam Guide
    bbg_exit_window: Filter
    psc_housing_entry_window: Filter
    nose: EllipticGuide

    # pulse shaping choppers (not in a named section ...)
    pulse_shaping_chopper_1: DiscChopper
    pulse_shaping_chopper_2: DiscChopper

    # curved guide section
    unit_3_curved: StraightGuides
    psc_exit_window: Filter
    psc_monitor: FissionChamber
    curved_entrance_window: Filter
    unit_4_curved: StraightGuides
    unit_4_exit_window: Filter
    frame_overlap_chopper_1: DiscChopper
    unit_5_entry_window: Filter
    unit_5_curved: StraightGuides
    unit_6_curved: StraightGuides
    unit_7_curved: StraightGuides
    unit_8_curved: StraightGuides
    unit_8_exit_window: Filter
    frame_overlap_chopper_2: DiscChopper
    unit_9_entry_window: Filter
    unit_9_curved: StraightGuides
    unit_10_curved: StraightGuides
    unit_11_curved: StraightGuides
    unit_12_curved: StraightGuides
    unit_13_curved: StraightGuides
    unit_14_curved: StraightGuides
    unit_15_curved: StraightGuides

    # expanding guide section
    unit_16_bw_insert: EllipticGuide  # Bunker wall insert, first part
    unit_17_bw_insert: EllipticGuide  # Bunker wall insert, second part
    unit_17_exit_window: Filter
    overlap_monitor: BeamCurrentMonitor
    unit_18_entry_window: Filter
    unit_18_expanding: EllipticGuide
    unit_19_expanding: EllipticGuide
    unit_20_expanding: EllipticGuide
    unit_21_expanding: EllipticGuide
    unit_22_expanding: EllipticGuide
    unit_23_expanding: EllipticGuide
    unit_24_expanding: EllipticGuide
    unit_25_expanding: EllipticGuide
    unit_26_expanding: EllipticGuide
    unit_27_expanding: EllipticGuide
    unit_28_expanding: EllipticGuide
    unit_28_exit_window: Filter

    # straight guide transport section
    unit_29_entry_window: Filter
    unit_29_straight: StraightGuide
    unit_30_straight: StraightGuide
    unit_31_straight: StraightGuide
    unit_32_straight: StraightGuide
    unit_33_straight: StraightGuide
    unit_34_straight: StraightGuide
    unit_35_straight: StraightGuide
    unit_36_straight: StraightGuide
    unit_37_straight: StraightGuide
    unit_38_straight: StraightGuide
    unit_39_straight: StraightGuide
    unit_40_straight: StraightGuide
    unit_41_straight: StraightGuide
    unit_42_straight: StraightGuide
    unit_43_straight: StraightGuide
    bandwidth_chopper_1: DiscChopper
    bandwidth_chopper_2: DiscChopper
    unit_43_exit_window: Filter
    bandwidth_monitor: BeamCurrentMonitor
    attenuator_1: Attenuator
    attenuator_2: Attenuator
    attenuator_3: Attenuator
    unit_44_entry_window: Filter
    #
    unit_44_straight: StraightGuide
    unit_45_straight: StraightGuide
    unit_46_straight: StraightGuide
    unit_47_straight: StraightGuide
    unit_48_straight: StraightGuide
    unit_49_straight: StraightGuide
    unit_50_straight: StraightGuide
    unit_51_straight: StraightGuide
    unit_52_straight: StraightGuide
    unit_53_straight: StraightGuide
    unit_54_straight: StraightGuide
    unit_55_straight: StraightGuide
    unit_56_straight: StraightGuide
    unit_57_straight: StraightGuide
    unit_58_straight: StraightGuide
    unit_59_straight: StraightGuide
    unit_60_straight: StraightGuide
    unit_61_straight: StraightGuide
    unit_62_straight: StraightGuide
    unit_63_straight: StraightGuide
    unit_64_straight: StraightGuide
    unit_65_straight: StraightGuide
    unit_66_straight: StraightGuide
    unit_67_straight: StraightGuide
    unit_68_straight: StraightGuide
    unit_69_straight: StraightGuide
    unit_70_straight: StraightGuide
    unit_71_straight: StraightGuide
    unit_72_straight: StraightGuide
    unit_73_straight: StraightGuide
    unit_74_straight: StraightGuide
    unit_75_straight: StraightGuide

    # focusing section
    unit_76_closing: EllipticGuide
    unit_77_closing: EllipticGuide
    unit_78_closing: EllipticGuide
    unit_79_closing: EllipticGuide
    unit_80_closing: EllipticGuide
    unit_81_closing: EllipticGuide
    unit_82_closing: EllipticGuide
    unit_83_closing: EllipticGuide
    unit_84_closing: EllipticGuide
    unit_85_closing: EllipticGuide
    jaw_3: Jaw
    unit_86_closing: EllipticGuide
    jaw_2: Jaw
    unit_87_closing: EllipticGuide
    jaw_1: Jaw
    unit_88_closing: EllipticGuide
    unit_88_exit_window: Filter
    mask: Slit
    normalization_monitor: GEM2D
    slit: Slit

    sample_origin: Component

    @classmethod
    @calibration
    def from_calibration(cls, parameters: dict):
        from .parameters import primary_parameters
        if len(parameters) == 0:
            parameters = primary_parameters()
        return super().from_calibration(parameters)


